import io
import json
import inspect

from dataclasses import dataclass
from collections import OrderedDict
from typing import Dict, List, Union
from ruamel.yaml import YAML

import kclvm
import kclvm.config
import kclvm.api.object as obj

from kclvm.api.object.internal import Undefined, UndefinedType
from kclvm.api.object.schema import (
    SETTINGS_OUTPUT_KEY,
    SCHEMA_SETTINGS_ATTR_NAME,
)


KCL_PLAN_TYPE = [
    obj.KCLObjectType.INTEGER,  # int
    obj.KCLObjectType.FLOAT,  # float
    obj.KCLObjectType.STRING,  # string
    obj.KCLObjectType.BOOLEAN,  # True, False
    obj.KCLObjectType.NUMBER_MULTIPLIER,  # 1M, 1Ki
    obj.KCLObjectType.NONE,  # None
    obj.KCLObjectType.UNDEFINED,  # Undefined
    obj.KCLObjectType.DICT,  # dict
    obj.KCLObjectType.LIST,  # list
    obj.KCLObjectType.SCHEMA,  # dict with __settings__
]
LIST_DICT_TEMP_KEY = "$"


def is_kcl_schema(value: dict):
    return (
        value is not None
        and isinstance(value, dict)
        and SCHEMA_SETTINGS_ATTR_NAME in value
        and SETTINGS_OUTPUT_KEY in value[SCHEMA_SETTINGS_ATTR_NAME]
    )


def order_dict(d: any) -> dict:
    result = {}
    for k, v in sorted(d.items()):
        if isinstance(v, dict):
            result[k] = order_dict(v)
        else:
            result[k] = v
    return result


def handle_schema(value: dict):
    # on kcl schema
    filtered = filter_results(value)
    if filtered is None:
        return filtered, False
    settings = SCHEMA_SETTINGS_ATTR_NAME
    output_type = SETTINGS_OUTPUT_KEY
    if settings in value and value[settings][output_type] == obj.SETTINGS_OUTPUT_IGNORE:
        if len(filtered) <= 1:
            return None, False
        else:
            return filtered[1:], True
    standalone = False
    if (
        settings in value
        and value[settings][output_type] == obj.SETTINGS_OUTPUT_STANDALONE
    ):
        standalone = True
    return filtered, standalone


def filter_results(keyvalues: Union[dict, list]) -> List[dict]:
    if keyvalues is None:
        return None

    # Plan list value with the yaml stream format.
    if isinstance(keyvalues, list):
        result = []
        for kv in keyvalues:
            result += filter_results(kv)
        return result

    # index 0 for in-line keyvalues output, index 1: for standalone keyvalues outputs
    results = [OrderedDict()]

    for key, value in keyvalues.items():
        if value is None and kclvm.config.disable_none:
            continue
        if isinstance(key, str) and key.startswith("_"):
            pass
        elif value is Undefined or isinstance(value, UndefinedType):
            pass
        elif inspect.isclass(value):
            pass
        elif inspect.isfunction(value):
            pass
        elif inspect.ismodule(value):
            pass
        elif inspect.isfunction(value):
            pass
        elif is_kcl_schema(value):
            filtered, standalone = handle_schema(value)
            if filtered is not None:
                if standalone:
                    # if the instance is marked as 'STANDALONE', treat it as a separate one and
                    # extend it and derived STANDALONE instances to results.
                    results.extend(filtered)
                else:
                    # else put it as the value of the key of results
                    if len(results) > 0 and len(filtered) > 0:
                        results[0][key] = filtered[0]
                    if len(filtered) > 1:
                        # if the value has derived 'STANDALONE' instances, extend them
                        results.extend(filtered[1:])
        elif isinstance(value, dict):
            filtered = filter_results(value)
            if len(results) > 0 and len(filtered) > 0:
                results[0][key] = filtered[0]
            if len(results) > 0 and len(filtered) > 1:
                # if the value has derived 'STANDALONE' instances, extend them
                results.extend(filtered[1:])
        elif isinstance(value, list):
            filtered_list = []
            standalone_list = []
            ignore_schema_count = 0
            for i in value:
                if is_kcl_schema(i):
                    filtered, standalone = handle_schema(i)
                    if filtered is None:
                        ignore_schema_count += 1
                        continue
                    if filtered:
                        if standalone:
                            standalone_list.extend(filtered)
                        else:
                            filtered_list.extend(filtered)
                elif isinstance(i, dict):
                    filtered = filter_results(i)
                    if filtered:
                        filtered_list.extend(filtered)
                elif i is None and kclvm.config.disable_none:
                    continue
                elif not isinstance(i, UndefinedType):
                    # Filter list elements
                    filtered = filter_results({LIST_DICT_TEMP_KEY: i})
                    if (
                        len(results) > 0
                        and len(filtered) > 0
                        and LIST_DICT_TEMP_KEY in filtered[0]
                    ):
                        filtered_list.append(filtered[0][LIST_DICT_TEMP_KEY])
                    if len(results) > 0 and len(filtered) > 1:
                        # if the value has derived 'STANDALONE' instances, extend them
                        results.extend(filtered[1:])
            schema_in_list_count = ignore_schema_count + len(standalone_list)
            if len(results) > 0 and len(value) == 0:
                results[0][key] = []
            if len(results) > 0 and 0 <= schema_in_list_count < len(value):
                results[0][key] = filtered_list
            if standalone_list:
                results.extend(standalone_list)
        else:
            results[0][key] = value
    return results


@dataclass
class ObjectPlanner:
    """Planner is used to modify VM exec result to YAML"""

    def __init__(
        self, *, sort_keys: bool = False, include_schema_type_path: bool = False
    ) -> None:
        self.sort_keys = sort_keys
        self.include_schema_type_path = include_schema_type_path

    def to_python(self, v):
        if isinstance(v, obj.KCLObject):
            if isinstance(v, obj.KCLLiteralObject):
                return v.value
            elif isinstance(v, obj.KCLSchemaObject):
                result = {_k: self.to_python(_v) for _k, _v in v.attrs.items()}
                if self.include_schema_type_path:
                    result["@type"] = v.full_type_str()
                return result
            elif isinstance(v, (obj.KCLListObject, obj.KCLTupleObject)):
                return [self.to_python(_v) for _v in v.value]
            elif isinstance(v, obj.KCLDictObject):
                return {_k: self.to_python(_v) for _k, _v in v.value.items()}
            elif isinstance(v, (obj.KCLUndefinedObject, obj.KCLFunctionObject)):
                return Undefined
            elif isinstance(v, obj.KCLNameConstantObject):
                return v.value
            else:
                return Undefined
        elif isinstance(v, (list, tuple, set)):
            return [self.to_python(_v) for _v in v]
        elif isinstance(v, dict):
            return {_k: self.to_python(_v) for _k, _v in v.items()}
        elif isinstance(v, (int, float, str, bool)) or v is None:
            return v
        elif v is Undefined or isinstance(v, UndefinedType):
            return v
        else:
            raise Exception("Invalid KCL Object")

    def plan_dict(self, var_dict: Dict[str, obj.KCLObject]) -> Dict[str, any]:
        return {
            k: self.to_python(v)
            for k, v in var_dict.items()
            if v and v.type() in KCL_PLAN_TYPE
        }

    def plan(
        self, var: Union[Dict[str, obj.KCLObject], list]
    ) -> Union[Dict[str, obj.KCLObject], list]:
        if isinstance(var, dict):
            return self.plan_dict(var)
        elif isinstance(var, list):
            return [self.plan_dict(v) for v in var]
        elif isinstance(var, obj.KCLListObject):
            return [self.to_python(v) for v in var.items]
        raise Exception(f"Invalid plan object type {type(var)}")


@dataclass
class YAMLPlanner(ObjectPlanner):
    def __init__(
        self, *, sort_keys: bool = False, include_schema_type_path: bool = False
    ) -> None:
        super().__init__(
            sort_keys=sort_keys, include_schema_type_path=include_schema_type_path
        )

    def plan(self, var_dict: Dict[str, obj.KCLObject], to_py: bool = not kclvm.config.is_target_native) -> str:
        plan_obj = super().plan(var_dict) if to_py else var_dict
        # Represent OrderedDict as dict.
        yaml = YAML()
        yaml.representer.add_representer(
            OrderedDict,
            lambda dumper, data: dumper.represent_mapping(
                "tag:yaml.org,2002:map", data
            ),
        )
        # Convert tuple to list.
        yaml.representer.add_representer(
            tuple,
            lambda dumper, data: dumper.represent_sequence(
                "tag:yaml.org,2002:seq", data
            ),
        )
        # Convert None to null
        yaml.representer.add_representer(
            type(None),
            lambda dumper, data: dumper.represent_scalar(
                "tag:yaml.org,2002:null", "null"
            ),
        )
        yaml.representer.add_representer(
            str,
            lambda dumper, data: dumper.represent_scalar(
                "tag:yaml.org,2002:str", data, style="|"
            )
            if "\n" in data
            else dumper.represent_str(data),
        )
        results = filter_results(plan_obj)
        if self.sort_keys:
            results = [order_dict(r) for r in results]
        outputs = ""
        with io.StringIO() as buf:
            if results:
                for result in results[:-1]:
                    if result:
                        yaml.dump(result, buf)
                        buf.write("---\n")
                if results[-1]:
                    yaml.dump(results[-1], buf)
                outputs = buf.getvalue()
        return outputs


class JSONPlanner(ObjectPlanner):
    def __init__(
        self, *, sort_keys: bool = False, include_schema_type_path: bool = False
    ) -> None:
        super().__init__(
            sort_keys=sort_keys, include_schema_type_path=include_schema_type_path
        )

    def plan(
        self,
        var: Union[Dict[str, obj.KCLObject], list],
        *,
        only_first=False,
        to_py: bool = True,
    ) -> str:
        plan_obj = super().plan(var) if to_py else var
        results = filter_results(plan_obj)
        if self.sort_keys:
            results = [order_dict(r) for r in results]
        results = [r for r in results if r]
        return (
            json.dumps(
                results if not only_first else results[0],
                default=lambda o: o.__dict__,
                indent=4,
            )
            if results
            else ""
        )


def plan(
    val_map: Dict[str, obj.KCLObject],
    *,
    sort_keys: bool = False,
    include_schema_type_path: bool = False,
) -> str:
    return YAMLPlanner(
        sort_keys=sort_keys, include_schema_type_path=include_schema_type_path
    ).plan(val_map)



# Copyright 2021 The KCL Authors. All rights reserved.

from typing import Optional

import kclvm.kcl.ast as ast
import kclvm.compiler.parser.parser as parser
import kclvm.vm as vm
import kclvm.vm.planner as planner
import kclvm.compiler.build.compiler as compiler

MAIN_PKG_NAME = "__main__"


def EvalCode(filename: str, code: Optional[str] = None) -> str:
    # Parser
    module = parser.ParseFile(filename, code)
    return EvalAST(module)


def EvalAST(module: ast.Module) -> str:
    module.pkg = MAIN_PKG_NAME
    # Compiler
    bytecode = compiler.CompileProgram(
        ast.Program(
            root=MAIN_PKG_NAME, main=MAIN_PKG_NAME, pkgs={MAIN_PKG_NAME: [module]}
        )
    )
    # VM run
    result = vm.Run(bytecode)
    # YAML plan
    return planner.YAMLPlanner().plan(result.filter_by_path_selector())



from ast import literal_eval
from copy import deepcopy
from pathlib import Path
from typing import List
import argparse as _argparse
import ruamel.yaml as _yaml
from io import StringIO

import kclvm.kcl.ast as ast
import kclvm.kcl.error as kcl
import kclvm.config

verbose = 0
"""
Print more information and intermediate representation.
"""

debug = 0
"""
Print debug information for developers.
"""

save_temps = False
"""
Save the intermediate representation files.
"""

output = None
"""
The name of the output file.
"""

input_file = None
"""
The name of the input file.
"""

current_path = None
"""
The current path where KCL is executed.
"""

disable_none = False
"""
Disable dumping values that are None.
"""

strict_range_check = False
"""
Whether perform strict numberic range check
"""

arguments = []
"""
Top level arguments.
"""

list_option_mode = 0
"""
List option mode (>= 0).
"""

disable_schema_check = False
"""
Disabel schema check
"""

options_help_message = ""
"""
Optopns help message
"""

path_selector = []
"""
The path selector
"""

overrides = []
"""
The configuration override path and value
"""

is_target_native = True
"""Whether the target is 'native'"""


def dump():
    if verbose:
        print("=== Configurations ===")
        print("verbose      = {}".format(verbose))
        print("save_temps   = {}".format(save_temps))
        print("output       = {}".format(output))
        print("disable_none = {}".format(disable_none))
        print("input_file   = {}".format(input_file))
        print("current_path = {}".format(current_path))
        print("arguments    = {}".format(arguments))
        print("strict_range_check = {}".format(strict_range_check))

        print("list_option_mode = {}".format(list_option_mode))
        print("disable_schema_check = {}".format(disable_schema_check))


def parse_config(argsdict: dict):
    """
    Parse kclvm config with cli argument dict
    """
    if not argsdict:
        return
    kclvm.config.verbose = (
        int(argsdict.get(KCLCLIFlag.VERBOSE, 0)) or kclvm.config.verbose
    )
    kclvm.config.debug = int(argsdict.get(KCLCLIFlag.DEBUG, 0)) or kclvm.config.debug
    kclvm.config.list_option_mode = (
        int(argsdict.get(KCLCLIFlag.LIST_OPTION_MODE, 0))
        or kclvm.config.list_option_mode
    )
    kclvm.config.save_temps = (
        argsdict.get(KCLCLIFlag.SAVE_TEMPS, False) or kclvm.config.save_temps
    )
    kclvm.config.strict_range_check = (
        argsdict.get(KCLCLIFlag.STRICT_RANGE_CHECK) or kclvm.config.strict_range_check
    )
    kclvm.config.disable_none = (
        argsdict.get(KCLCLIFlag.DISABLE_NONE, False) or kclvm.config.disable_none
    )
    kclvm.config.path_selector = argsdict.get(
        KCLCLIFlag.PATH_SELECTOR, kclvm.config.path_selector
    )
    kclvm.config.overrides = argsdict.get(KCLCLIFlag.OVERRIDES, kclvm.config.overrides)
    kclvm.config.input_file = (
        argsdict.get(KCLCLIFlag.FILES)
        or argsdict.get(KCLCLIFlag.FILE)
        or kclvm.config.input_file
    )
    kclvm.config.output = (
        argsdict.get(KCLCLIFlag.OUTPUT)
        if argsdict.get(KCLCLIFlag.OUTPUT)
        else kclvm.config.output
    )
    kclvm.config.arguments += argsdict.get(KCLCLIFlag.ARGUMENT, [])


def _deal_key_value(values, keys, vals):
    is_valid_value = False
    _k, _v = None, None
    try:
        _k, _v = values
        _v = literal_eval(_v)
    except Exception:
        pass
    # _v is a string type, and avoid quotation ' and " mark escape error
    _v = _v.replace("'", "\\'").replace('"', '\\"') if isinstance(_v, str) else _v
    # the _v can only be KCL internal type
    if _v is None or isinstance(_v, (bool, int, float, str, list, dict)):
        is_valid_value = True
    if _k is None or not isinstance(_k, str) or _k.strip() == "":
        kcl.report_exception(
            err_type=kcl.ErrType.IllegalArgumentError_TYPE,
            arg_msg=f"Invalid option name: '{_k}'. should be a non-empty string",
        )
    if not is_valid_value:
        kcl.report_exception(
            err_type=kcl.ErrType.IllegalArgumentError_TYPE,
            arg_msg=f"Invalid option value: '{_object_to_yaml_str(_v)}' for option(\"{_k}\").",
        )
    keys.append(_k)
    vals.append(_v)


class KCLCLIFlag:
    FILE = "file"
    FILES = "files"
    DISABLE_NONE = "disable_none"
    DEBUG = "debug"
    STRICT_RANGE_CHECK = "strict_range_check"
    OUTPUT = "output"
    VERBOSE = "verbose"
    SAVE_TEMPS = "save_temps"
    ARGUMENT = "argument"
    PATH_SELECTOR = "path_selector"
    OVERRIDES = "overrides"
    LIST_OPTION_MODE = "list_option_mode"


class KCLTopLevelArgumentAction(_argparse.Action):
    """KCL CLI top level argument argparse action"""

    def __call__(self, parser, namespace, values, option_string=None):
        invalid_arg_msg = 'Invalid value for option "--argument(-D)"'
        split_values = values.split("=")
        _keys, _vals = [], []
        if len(split_values) == 2 and str(split_values[1]).strip() != "":
            try:
                _deal_key_value(split_values, _keys, _vals)
            except Exception as err:
                err_msg = err.arg_msg if isinstance(err, kcl.KCLException) else err
                kcl.report_exception(
                    err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                    arg_msg=f"{invalid_arg_msg}: {err_msg}",
                )
        else:
            kcl.report_exception(
                err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                arg_msg=f"{invalid_arg_msg}: should be in <name>=<value> pattern, got: {values}",
            )
        args = getattr(namespace, self.dest)
        args = [] if args is None else deepcopy(args)
        args += [(_k, _v) for _k, _v in zip(_keys, _vals)]
        setattr(namespace, self.dest, args)


class KCLPathSelectorAction(_argparse.Action):
    """KCL path selector action"""

    def __call__(self, parser, namespace, values, option_string=None):
        def report_exception():
            kcl.report_exception(
                err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                arg_msg=f"Invalid value for option \"--path-selector(-S)\": '{values}'. path selector should be in <pkg>:<identifier> pattern",
            )

        split_values = values.split(":", 1)
        if len(split_values) > 2:
            report_exception()
        split_values = split_values if len(split_values) == 2 else ["", *split_values]
        args = getattr(namespace, self.dest)
        args = [] if args is None else deepcopy(args)
        args += [split_values]
        setattr(namespace, self.dest, args)


class KCLOverrideAction(_argparse.Action):
    """KCL path override action"""

    def __call__(self, parser, namespace, values, option_string=None):
        def report_exception():
            kcl.report_exception(
                err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                arg_msg=f"Invalid value for option \"--override(-O)\": '{values}'. the override expr should be in <the variable path selector>=<value> pattern",
            )

        args = getattr(namespace, self.dest)
        if "=" in values:
            split_values = values.split("=", 1)
            paths = split_values[0].split(":", 1)
            if len(split_values) < 2 or len(paths) > 2:
                report_exception()
            paths.append(split_values[1])
            args = [] if args is None else deepcopy(args)
            paths = paths if len(paths) == 3 else ["", *paths]
            paths += [ast.OverrideAction.CREATE_OR_UPDATE]
            args += [paths]
        elif values.endswith("-"):
            paths = values[:-1].split(":", 1)
            if len(paths) > 2:
                report_exception()
            paths = paths if len(paths) == 2 else ["", *paths]
            paths += ["", ast.OverrideAction.DELETE]
            args += [paths]
        setattr(namespace, self.dest, args)


class KCLCLISettingAction:
    """KCL CLI top level setting argparse action"""

    DEFAULT_SETTING_PATH = "kcl.yaml"
    OPTION_FILE_TYPES = [".yaml", ".yml"]
    KCL_OPTION_KEY = "kcl_options"
    KCL_CLI_CONFIG_KEY = "kcl_cli_configs"

    ARGUMENTS_ITEM_KEY = "key"
    ARGUMENTS_ITEM_VALUE = "value"

    KCL_OPTION_EXAMPLE = """
kcl_options:
  - key: myArg # the option key must be a string value
    value: myArgValue"""
    INVALID_OPTION_MSG = f"invalid kcl_options value, should be list of key/value mapping. \n=== A good example will be:==={KCL_OPTION_EXAMPLE}"

    def deal_config_obj(self, data: dict, keys: list, vals: list) -> bool:
        if data is None or not isinstance(data, dict):
            kcl.report_exception(
                err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                arg_msg=f"setting file content should be a mapping, got: {_object_to_yaml_str(data)}",
            )
        # Deal Arguments
        kcl_options_mapping_list = data.get(self.KCL_OPTION_KEY) or []
        if not isinstance(kcl_options_mapping_list, list):
            kcl.report_exception(
                err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                arg_msg=f"{self.INVALID_OPTION_MSG}\n=== got: ===\n{_object_to_yaml_str({self.KCL_OPTION_KEY: kcl_options_mapping_list})}",
            )
        for key_value in kcl_options_mapping_list:
            if key_value is None or not isinstance(key_value, dict):
                kcl.report_exception(
                    err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                    arg_msg=f"{self.INVALID_OPTION_MSG}\n=== got: ===\n{_object_to_yaml_str({self.KCL_OPTION_KEY: [key_value]})}",
                )
            _deal_key_value(
                (
                    key_value.get(self.ARGUMENTS_ITEM_KEY),
                    key_value.get(self.ARGUMENTS_ITEM_VALUE),
                ),
                keys,
                vals,
            )
        # Deal KCL CLI parameters
        data = data.get(self.KCL_CLI_CONFIG_KEY, {})
        parse_config(data)
        return len(data.get(KCLCLIFlag.FILES) or data.get(KCLCLIFlag.FILE) or []) > 0

    def deal_setting_file(self, filename, keys, vals):
        data = None
        if any(filename.endswith(t) for t in self.OPTION_FILE_TYPES):
            try:
                data = _yaml.safe_load(Path(filename).read_text(encoding="utf-8"))
            except Exception as err:
                err_msg = err.arg_msg if isinstance(err, kcl.KCLException) else err
                kcl.report_exception(
                    err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                    file_msgs=[kcl.ErrFileMsg(filename=filename)],
                    arg_msg=f"Invalid yaml content of setting file:\n{err_msg}",
                )
            try:
                return self.deal_config_obj(data or {}, keys, vals)
            except Exception as err:
                err_msg = err.arg_msg if isinstance(err, kcl.KCLException) else err
                kcl.report_exception(
                    err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                    file_msgs=[kcl.ErrFileMsg(filename=filename)],
                    arg_msg=f"Invalid configuration in setting file:\n{err_msg}",
                )
        else:
            kcl.report_exception(
                err_type=kcl.ErrType.IllegalArgumentError_TYPE,
                file_msgs=[kcl.ErrFileMsg(filename=filename)],
                arg_msg='Invalid value for option "--setting(-Y)": the setting file should be in yaml format',
            )

    def deal(self, filenames: List[str]) -> list:
        work_dir = ""
        if not filenames and not Path(self.DEFAULT_SETTING_PATH).exists():
            return [], work_dir
        args = []
        for filename in filenames or [self.DEFAULT_SETTING_PATH]:
            _keys, _vals = [], []
            if self.deal_setting_file(filename, _keys, _vals):
                work_dir = str(Path(filename).parent)
            args += [(_k, _v) for _k, _v in zip(_keys, _vals)]
        return args, work_dir


def _object_to_yaml_str(obj, options=None):
    if not isinstance(obj, (list, dict)):
        return obj
    yaml_config = _yaml.YAML()
    yaml_config.indent(mapping=3, sequence=2, offset=0)
    yaml_config.allow_duplicate_keys = True

    # show null
    def my_represent_none(self, data):
        return self.represent_scalar("tag:yaml.org,2002:null", "null")

    yaml_config.representer.add_representer(type(None), my_represent_none)

    # yaml to string
    if options is None:
        options = {}
    string_stream = StringIO()
    yaml_config.dump(obj, string_stream, **options)
    output_str = string_stream.getvalue()
    string_stream.close()
    return output_str



# Copyright 2021 The KCL Authors. All rights reserved.

from .config import (
    verbose,
    debug,
    save_temps,
    output,
    input_file,
    current_path,
    disable_none,
    strict_range_check,
    arguments,
    list_option_mode,
    disable_schema_check,
    options_help_message,
    path_selector,
    overrides,
    dump,
    parse_config,
    is_target_native,
    KCLTopLevelArgumentAction,
    KCLPathSelectorAction,
    KCLOverrideAction,
    KCLCLISettingAction,
)
from .modfile_pb2 import KclModFile
from .settings import load_settings_files

__all__ = [
    "verbose",
    "debug",
    "save_temps",
    "output",
    "input_file",
    "current_path",
    "disable_none",
    "strict_range_check",
    "arguments",
    "list_option_mode",
    "disable_schema_check",
    "options_help_message",
    "path_selector",
    "overrides",
    "dump",
    "parse_config",
    "is_target_native",
    "KCLTopLevelArgumentAction",
    "KCLPathSelectorAction",
    "KCLOverrideAction",
    "KCLCLISettingAction",
    "KclModFile",
    "load_settings_files",
]



"""basic checker for Python code"""
import re
from typing import Pattern

import kclvm.tools.lint.message.message as message
import kclvm.kcl.ast as ast
import kclvm.tools.lint.lint.utils as utils
from kclvm.tools.lint.checkers.base_checker import BaseChecker


MSGS = {
    "C0103": (
        '%s name "%s" doesn\'t conform to %s.',
        "Invalid-name.",
        "{0} name {1} doesn't conform to {2}.",
    ),
    "C0104": (
        'Disallowed name "%s".',
        "Disallowed-name.",
        'Disallowed name "{0}."',
    ),
}


class NamingStyle:
    ANY: Pattern[str] = re.compile(".*")
    MOD_NAME_RGX: Pattern[str] = ANY
    PKG_NAME_RGX: Pattern[str] = ANY
    SCHEMA_NAME_RGX: Pattern[str] = ANY
    MIXIN_NAME_RGX: Pattern[str] = ANY
    PROTOCOL_NAME_RGX: Pattern[str] = ANY
    SCHEMA_ATTRIBUTE_RGX: Pattern[str] = ANY
    VARIABLE_RGX: Pattern[str] = ANY
    ARGUMENT_RGX: Pattern[str] = ANY
    DEFAULT_NAME_RGX: Pattern[str] = ANY

    @classmethod
    def get_regex(cls, name_type):
        return {
            "module": cls.MOD_NAME_RGX,
            "package": cls.PKG_NAME_RGX,
            "schema": cls.SCHEMA_NAME_RGX,
            "mixin": cls.MIXIN_NAME_RGX,
            "protocol": cls.PROTOCOL_NAME_RGX,
            "argument": cls.ARGUMENT_RGX,
            "variable": cls.VARIABLE_RGX,
            "schema_attribute": cls.SCHEMA_ATTRIBUTE_RGX,
        }[name_type]


class SnakeCaseStyle(NamingStyle):
    ANY: Pattern[str] = re.compile(".*")
    MOD_NAME_RGX: Pattern[str] = re.compile(r"^\$?[a-z][a-z_]*$")
    PKG_NAME_RGX: Pattern[str] = re.compile(r"^\$?[a-z][a-z_]*$")
    SCHEMA_NAME_RGX: Pattern[str] = ANY
    MIXIN_NAME_RGX: Pattern[str] = ANY
    PROTOCOL_NAME_RGX: Pattern[str] = ANY
    VARIABLE_RGX: Pattern[str] = ANY
    SCHEMA_ATTRIBUTE_RGX: Pattern[str] = ANY
    ARGUMENT_RGX: Pattern[str] = ANY
    DEFAULT_NAME_RGX: Pattern[str] = ANY


class CamelCaseStyle(NamingStyle):
    """Regex rules for camelCase naming style."""

    ANY: Pattern[str] = re.compile(".*")
    MOD_NAME_RGX: Pattern[str] = ANY
    PKG_NAME_RGX: Pattern[str] = ANY
    SCHEMA_NAME_RGX: Pattern[str] = ANY
    MIXIN_NAME_RGX: Pattern[str] = ANY
    PROTOCOL_NAME_RGX: Pattern[str] = ANY
    VARIABLE_RGX: Pattern[str] = ANY
    SCHEMA_ATTRIBUTE_RGX: Pattern[str] = re.compile(r"^\$?[a-z][a-zA-Z]*$")
    ARGUMENT_RGX: Pattern[str] = re.compile(r"^\$?[a-z][a-zA-Z]*$")
    DEFAULT_NAME_RGX: Pattern[str] = re.compile(r"^\$?[a-z][a-zA-Z]*$")


class PascalCaseStyle(NamingStyle):
    """Regex rules for PascalCase naming style."""

    ANY: Pattern[str] = re.compile(".*")
    MOD_NAME_RGX: Pattern[str] = ANY
    PKG_NAME_RGX: Pattern[str] = ANY
    SCHEMA_NAME_RGX: Pattern[str] = re.compile(r"^\$?[A-Z][a-zA-Z\d]*$")
    MIXIN_NAME_RGX: Pattern[str] = re.compile(r"^\$?[A-Z][a-zA-Z\d]*Mixin$")
    PROTOCOL_NAME_RGX: Pattern[str] = re.compile(r"^\$?[A-Z][a-zA-Z\d]*Protocol$")
    VARIABLE_RGX: Pattern[str] = ANY
    SCHEMA_ATTRIBUTE_RGX: Pattern[str] = ANY
    ARGUMENT_RGX: Pattern[str] = ANY
    DEFAULT_NAME_RGX: Pattern[str] = ANY


class UpperCaseStyle(NamingStyle):
    """Regex rules for UPPER_CASE naming style."""

    ANY: Pattern[str] = re.compile(".*")
    MOD_NAME_RGX: Pattern[str] = ANY
    PKG_NAME_RGX: Pattern[str] = ANY
    SCHEMA_NAME_RGX: Pattern[str] = ANY
    MIXIN_NAME_RGX: Pattern[str] = ANY
    PROTOCOL_NAME_RGX: Pattern[str] = ANY
    VARIABLE_RGX: Pattern[str] = ANY
    SCHEMA_ATTRIBUTE_RGX: Pattern[str] = ANY
    ARGUMENT_RGX: Pattern[str] = ANY
    DEFAULT_NAME_RGX: Pattern[str] = re.compile(r"^\$?[^\W\da-z_][^\Wa-z]*$")


class AnyStyle(NamingStyle):
    pass


NAMING_STYLES = {
    "snake_case": SnakeCaseStyle,
    "camelCase": CamelCaseStyle,
    "PascalCase": PascalCaseStyle,
    "UPPER_CASE": UpperCaseStyle,
    "ANY": AnyStyle,
}


KNOWN_NAME_TYPES = {
    "module",
    "package",
    "schema",
    "mixin",
    "protocol",
    "argument",
    "variable",
    "schema_attribute",
}


class BasicChecker(BaseChecker):
    def __init__(self, linter) -> None:
        super().__init__(linter)
        self.name = "BaseCheck"
        self.code = None
        self.MSGS = MSGS
        self.module = None
        self.naming_rules = None
        self.bad_names = None
        self.prog = None

    def reset(self) -> None:
        self.msgs.clear()
        self.code = None
        self.module = None
        self.naming_rules = None
        self.bad_names = None

    def get_module(self, prog: ast.Program, code: str) -> None:
        assert isinstance(prog, ast.Program)
        assert code is not None
        self.prog = prog
        self.module = prog.pkgs["__main__"][0]
        self.code = code
        self.naming_rules = self._create_naming_rules()
        self.bad_names = self.options.bad_names

    def check(self, prog: ast.Program, code: str) -> None:
        assert isinstance(prog, ast.Program)
        assert code is not None
        self.reset()
        self.get_module(prog, code)
        self.walk(self.module)

    def _create_naming_rules(self):
        regexps = {}
        for name_type in KNOWN_NAME_TYPES:
            # naming rgx in config, .e.g  module-rgx : [^\W\dA-Z][^\WA-Z]+$
            custom_regex_setting_name = f"{name_type}_rgx"
            if getattr(self.options, custom_regex_setting_name):
                regexps[name_type] = re.compile(
                    getattr(self.options, custom_regex_setting_name)
                )
            else:
                # naming rgx in config, .e.g  module-rgx : [^\W\dA-Z][^\WA-Z]+$
                naming_style_name = f"{name_type}_naming_style"
                regexps[name_type] = NAMING_STYLES[
                    getattr(self.options, naming_style_name)
                ].get_regex(name_type)
        return regexps

    def _check_name(self, name: str, name_type: str):
        return self.naming_rules[name_type].search(name)

    def _disallowed_name(self, name: str):
        return name in self.bad_names

    def _get_name_style_or_rgx(self, name_type):
        name_rgx = f"{name_type}_rgx"
        name_style = f"{name_type}_naming_style"
        if getattr(self.options, name_rgx):
            return getattr(self.options, name_rgx)
        else:
            return getattr(self.options, name_style) + " naming style"

    def walk_RuleStmt(self, t: ast.RuleStmt) -> None:
        assert isinstance(t, ast.RuleStmt)
        if not self._check_name(t.name, "schema"):
            name_style_or_rgx = self._get_name_style_or_rgx("schema")
            self.msgs.append(
                message.Message(
                    "C0103",
                    self.module.filename,
                    MSGS["C0103"][0] % ("Schema", t.name, name_style_or_rgx),
                    utils.get_source_code(
                        self.module.filename, t.name_node.line, self.code
                    ),
                    (t.name_node.line, t.name_node.column),
                    ["Schema", t.name, name_style_or_rgx],
                )
            )
        self.generic_walk(t)

    def walk_SchemaStmt(self, t: ast.SchemaStmt) -> None:
        assert isinstance(t, ast.SchemaStmt)
        if t.is_mixin:
            if not self._check_name(t.name, "mixin"):
                name_style_or_rgx = self._get_name_style_or_rgx("mixin")
                self.msgs.append(
                    message.Message(
                        "C0103",
                        self.module.filename,
                        MSGS["C0103"][0] % ("Mixin", t.name, name_style_or_rgx),
                        utils.get_source_code(
                            self.module.filename, t.name_node.line, self.code
                        ),
                        (t.name_node.line, t.name_node.column),
                        ["Mixin", t.name, name_style_or_rgx],
                    )
                )
        elif t.is_protocol:
            if not self._check_name(t.name, "protocol"):
                name_style_or_rgx = self._get_name_style_or_rgx("schema")
                self.msgs.append(
                    message.Message(
                        "C0103",
                        self.module.filename,
                        MSGS["C0103"][0] % ("Protocol", t.name, name_style_or_rgx),
                        utils.get_source_code(
                            self.module.filename, t.name_node.line, self.code
                        ),
                        (t.name_node.line, t.name_node.column),
                        ["Protocol", t.name, name_style_or_rgx],
                    )
                )
        else:
            if not self._check_name(t.name, "schema"):
                name_style_or_rgx = self._get_name_style_or_rgx("schema")
                self.msgs.append(
                    message.Message(
                        "C0103",
                        self.module.filename,
                        MSGS["C0103"][0] % ("Schema", t.name, name_style_or_rgx),
                        utils.get_source_code(
                            self.module.filename, t.name_node.line, self.code
                        ),
                        (t.name_node.line, t.name_node.column),
                        ["Schema", t.name, name_style_or_rgx],
                    )
                )
        self.generic_walk(t)

    def walk_Arguments(self, t: ast.Arguments) -> None:
        assert isinstance(t, ast.Arguments)
        for arg in t.args:
            for name in arg.name_nodes:
                if not self._check_name(name.value, "argument"):
                    name_style_or_rgx = self._get_name_style_or_rgx("argument")
                    self.msgs.append(
                        message.Message(
                            "C0103",
                            self.module.filename,
                            MSGS["C0103"][0]
                            % ("Argument", name.value, name_style_or_rgx),
                            utils.get_source_code(
                                self.module.filename, t.line, self.code
                            ),
                            (t.line, t.column),
                            ["Argument", name.value, name_style_or_rgx],
                        )
                    )
        self.generic_walk(t)

    def walk_SchemaAttr(self, t: ast.SchemaAttr) -> None:
        assert isinstance(t, ast.SchemaAttr)
        if not self._check_name(t.name, "schema_attribute"):
            name_style_or_rgx = self._get_name_style_or_rgx("schema_attribute")
            self.msgs.append(
                message.Message(
                    "C0103",
                    self.module.filename,
                    MSGS["C0103"][0] % ("Schema attribute", t.name, name_style_or_rgx),
                    utils.get_source_code(self.module.filename, t.line, self.code),
                    (t.line, t.column),
                    ["Schema attribute", t.name, name_style_or_rgx],
                )
            )
        self.generic_walk(t)

    def walk_AssignStmt(self, t: ast.AssignStmt) -> None:
        assert isinstance(t, ast.AssignStmt)
        for target in t.targets:
            for name in target.name_nodes:
                if not self._check_name(name.value, "variable"):
                    name_style_or_rgx = self._get_name_style_or_rgx("variable")
                    self.msgs.append(
                        message.Message(
                            "C0103",
                            self.module.filename,
                            MSGS["C0103"][0]
                            % ("Variable", name.value, name_style_or_rgx),
                            utils.get_source_code(
                                self.module.filename, t.line, self.code
                            ),
                            (name.line, name.column),
                            ["Variable", name.value, name_style_or_rgx],
                        )
                    )
        self.generic_walk(t)

    def walk_Name(self, t: ast.Name) -> None:
        assert isinstance(t, ast.Name)
        if self._disallowed_name(t.value):
            self.msgs.append(
                message.Message(
                    "C0104",
                    self.module.filename,
                    MSGS["C0104"][0] % t.value,
                    utils.get_source_code(self.module.filename, t.line, self.code),
                    (t.line, t.column),
                    [t.value],
                )
            )
